"""Interwell connectivity through geometric considerations."""

from __future__ import annotations

import _core
import numpy as np
import pandas as pd
import scipy.linalg as sl
from numpy import ndarray

idx = pd.IndexSlice


def calc_gains_homogeneous(locations: pd.DataFrame, x_e: float, y_e: float) -> pd.DataFrame:
    r"""Calculate gains from injectors to producers using multiwell productivity index.

    The equation for the influence of injection on production is

    .. math::
        \mathbf{\Lambda} = \frac{\mathbf{A}_p^{-1}}{\sum \mathbf{A}_p^{-1}}
            \times \left(\mathbf{1} \times \mathbf{A}_p^{-1} \times \mathbf{A}_c^T - 1\right)
            - \left( \mathbf{A}_p^{-1} \times \mathbf{A}_c^T \right)

    Args
    ----------
    locations : pd.DataFrame
        columns include:
        `X` : x-location for well
        `Y` : y-location for well
        `Type`: ("Producer" or "Injector")
    x_e : float
        scaling in x-direction (unit size)
    y_e : float
        scaling in y-direction (unit size)

    Returns
    -------
    Lambda : pd.DataFrame
        contains the production-injection multiwell productivity index

    Notes
    -----
    This assumes a roughly rectangular unit with major axes at x and y. You might want
    to rotate your locations.

    Lambda is negative. Not sure why. Negate it to get positive injection leading to
    positive production.

    References
    ----------
    Kaviani, D. and Valkó, P.P., 2010. Inferring interwell connectivity using \
    multiwell productivity index (MPI). Journal of Petroleum Science and Engineering, \
    73(1-2), p.48-58.
    """
    locations = locations.copy()
    locations[["X", "Y"]] /= x_e
    y_D = y_e / x_e
    A_prod = calc_influence_matrix(locations, y_D, "prod")
    A_conn = calc_influence_matrix(locations, y_D, "conn")
    A_prod_inv = sl.inv(A_prod.to_numpy())
    term1 = A_prod_inv / np.sum(A_prod_inv)
    term2 = np.ones_like(A_prod_inv) @ A_prod_inv @ A_conn.to_numpy() - 1
    term3 = A_prod_inv @ A_conn.to_numpy()
    Lambda = term1 @ term2 - term3
    connectivity_df = pd.DataFrame(Lambda, index=A_prod.index, columns=A_conn.columns)
    return connectivity_df.rename_axis(index="Producers", columns="Injectors")


def translate_locations(
    locations: pd.DataFrame, x_col: str, y_col: str, type_col: str
) -> pd.DataFrame:
    """Translate locations  to prepare for building connectivity matrix.

    Moves the lower left edge of the reservoir to (0, 0), and sets up the matrix columns
    to work with `calc_gains_homogeneous`

    Args
    ----------
    locations : pd.DataFrame
        Has the x,y locations and type assignment for all wells
    x_col : str
        Column in `locations` for the x-location
    y_col : str
        Column in `locations` for the y-location
    type_col : str
        Column in `locations` for the type (producer or injector)

    Returns
    -------
    locations_out: pd.DataFrame
        columns are `X` (x-location), `Y` (y-location), `Type` (producer or injector)
    """
    locations_out = pd.DataFrame(index=locations.index, columns=["X", "Y", "Type"])
    locations_out["X"] = locations[x_col] - locations[x_col].min()
    locations_out["Y"] = locations[y_col] - locations[y_col].min()
    locations_out["Type"] = locations[type_col]
    return locations_out


def calc_influence_matrix(
    locations: pd.DataFrame, y_D: float, matrix_type: str = "conn", m_max: int = 100
) -> pd.DataFrame:
    """Calculate influence matrix A.

    Args
    ----------
    locations : pd.DataFrame
        Has the x,y locations and type assignment for all wells
        columns are `X` (x-location), `Y` (y-location), `Type` (producer or injector)
    y_D : float
        dimensionless scaling for y-direction
    matrix_type : str, choice of `conn` or `prod`
        injector-producer matrix or producer-producer matrix
    m_max : int > 0
        number of terms in the series to calculate. 100 is a good default.

    Returns
    -------
    influence_matrix : pd.DataFrame
        a matrix with the influences between wells
    """
    if matrix_type not in ["conn", "prod"]:
        msg = "matrix_type must be either `conn` or `prod`"
        raise ValueError(msg)
    XA = locations[locations.Type == "Producer"]
    XB = XA.copy() if matrix_type == "prod" else locations[locations.Type == "Injector"]
    influence_matrix = pd.DataFrame(
        index=pd.MultiIndex.from_product([XA.index, XB.index]), columns=["A"]
    )
    m = np.arange(1, m_max + 1, dtype="uint64")  # elements of sum
    for i, j in influence_matrix.index:
        x_i, y_i = XA.loc[i, ["X", "Y"]]
        x_j, y_j = XB.loc[j, ["X", "Y"]] + 1e-6
        influence_matrix.loc[idx[i, j], "A"] = calc_A_ij(x_i, y_i, x_j, y_j, y_D, m)
    return influence_matrix["A"].unstack().astype("float64")


def calc_A_ij(x_i: float, y_i: float, x_j: float, y_j: float, y_D: float, m: ndarray) -> float:
    r"""Calculate element in the influence matrix.

    .. math::
        A_{ij} = 2 \pi y_D (\frac13 - \frac{y_i}{y_D} +
            \frac{y_i^2 + y_j^2}{2 y_D^2})
            + \sum_{m=1}^\infty \frac{t_m}m \cos(m\pi \tilde x_i)
            \cos(m \pi \tilde x_j)
    where

    .. math::
        t_m = \frac{\cosh\left(m\pi (y_D - |\tilde y_i - \tilde y_j|)\right)
        + \cosh\left(m\pi (y_D - \tilde y_i - \tilde y_j\right)}
        {\sinh\left(m\pi y_D \right)}

    Args
    ----
    x_i : float
        x-location of i'th well
    y_i : float
        y-location of i'th well
    x_j : float
        x-location of j'th well
    y_j : float
        y-location of j'th well
    y_D : float
        dimensionless parameter for y-direction
    m : ndarray
        series terms, from 1 to m_max

    Returns
    -------
    A_ij : float

    References
    ----------
    Kaviani, D. and Valkó, P.P., 2010. Inferring interwell connectivity using \
    multiwell productivity index (MPI). Journal of Petroleum Science and Engineering, \
    73(1-2), p.48-58. https://doi.org/10.1016/j.petrol.2010.05.006
    """
    # Symmetry properties, see https://doi.org/10.1016/j.petrol.2010.05.006, A5-A6
    y_eD = y_D
    x_D = max([x_i, x_j])
    y_D = max([y_i, y_j])
    x_wD = min([x_i, x_j])
    y_wD = min([y_i, y_j])
    if not ((x_D - x_wD) > (y_D - y_wD)):
        y_eD = 1.0 / y_eD
    return _core.calc_A_ij(x_D, y_D, x_wD, y_wD, y_eD, m)



from __future__ import annotations

import numpy as np
import pandas as pd
import pytest
from numpy.typing import NDArray
from pywaterflood.multiwellproductivity import (
    calc_A_ij,
    calc_gains_homogeneous,
    calc_influence_matrix,
    translate_locations,
)

raw_locs = pd.DataFrame()


@pytest.fixture()
def locations():
    "Well locations for an inverted 5-spot."
    xy = [(1, 1), (20, 1), (1, 20), (20, 20), (10.5, 10.5)]
    x = np.array([x for x, _ in xy]) + np.random.normal(0, 0.5, len(xy))
    y = np.array([y for _, y in xy]) + np.random.normal(0, 0.5, len(xy))
    locations = pd.DataFrame(
        {
            "x": x,
            "y": y,
            "Type": ["Producer"] * 4 + ["Injector"],
        }
    )
    return translate_locations(locations, "x", "y", "Type")


def test_translate(locations):
    locations = translate_locations(locations, "X", "Y", "Type")
    assert locations["X"].min() <= 1e-10
    assert locations["Y"].min() <= 1e-10


def test_Aij():
    """From the worked example by Valkó et al, 2000.

    Development and Application of the Multiwell Productivity Index (MPI)"""
    m = 1 + np.arange(300, dtype="uint64")
    results = calc_A_ij(0.233351184, 0.36666667, 0.23333333, 0.36666667, 0.5, m)
    assert pytest.approx(10.6867, abs=0.01) == results
    for x_i in [0.0, 0.2, 0.3, 0.5]:
        old_A_ij = calc_A_ij_old(x_i, 0.36666667, 0.23333333, 0.36666667, 0.5, m)
        rust_A_ij = calc_A_ij(x_i, 0.36666667, 0.23333333, 0.36666667, 0.5, m)
        assert pytest.approx(old_A_ij, 1e-3) == rust_A_ij


def test_Aij_symmetry():
    idx = pd.IndexSlice
    m_max = 100
    locations = pd.DataFrame({"X": [0.2, 0.8], "Y": [0.3, 0.7]})
    y_D = 1.1
    influence_matrix = pd.DataFrame(
        index=pd.MultiIndex.from_product([locations.index, locations.index]), columns=["A"]
    )
    m = np.arange(1, m_max + 1, dtype="uint64")  # elements of sum
    for i, j in influence_matrix.index:
        x_i, y_i = locations.loc[i, ["X", "Y"]]
        x_j, y_j = locations.loc[j, ["X", "Y"]] + 1e-6
        influence_matrix.loc[idx[i, j], "A"] = calc_A_ij(x_i, y_i, x_j, y_j, y_D, m)
    influence_matrix = influence_matrix["A"].unstack().astype("float64")
    assert pytest.approx(influence_matrix.iloc[0, 1], 1e-4) == influence_matrix.iloc[1, 0]


def test_calc_gains(locations):
    """See that calc_gains works."""
    x_e = 34
    y_e = 22
    gains = calc_gains_homogeneous(locations, x_e, y_e)
    assert np.all(gains <= 0), "Gains must be negative"
    assert np.all(gains >= -1), "Gains can't be stronger than one"


def test_influence_matrix(locations):
    locations["X"] /= 20
    locations["Y"] /= 20
    matrix_conn = calc_influence_matrix(locations, y_D=0.7, matrix_type="conn", m_max=100)
    assert not np.isnan(matrix_conn.values).flatten().any()

    matrix_prod = calc_influence_matrix(locations, y_D=0.7, matrix_type="prod", m_max=100)
    assert not np.isnan(matrix_prod.values).flatten().any()
    with pytest.raises(ValueError, match="matrix_type must be"):
        calc_influence_matrix(locations, y_D=0.7, matrix_type="wrong")


# for regression testing
def calc_A_ij_old(x_i: float, y_i: float, x_j: float, y_j: float, y_D: float, m: NDArray) -> float:
    r"""Calculate element in the influence matrix.

    .. math::

        A_{ij} = 2 \pi y_D (\frac13 - \frac{y_i}{y_D} +
            \frac{y_i^2 + y_j^2}{2 y_D^2})
            + \sum_{m=1}^\infty \frac{t_m}m \cos(m\pi \tilde x_i)
            \cos(m \pi \tilde x_j)
    where

    .. math::
        t_m = \frac{\cosh\left(m\pi (y_D - |\tilde y_i - \tilde y_j|)\right)
        + \cosh\left(m\pi (y_D - \tilde y_i - \tilde y_j\right)}
        {\sinh\left(m\pi y_D \right)}

    Args
    ----
    x_i : float
        x-location of i'th well
    y_i : float
        y-location of i'th well
    x_j : float
        x-location of j'th well
    y_j : float
        y-location of j'th well
    y_D : float
        dimensionless parameter for y-direction
    m : ndarray
        series terms, from 1 to m_max

    Returns
    -------
    A_ij : float
    """
    y_j, y_i = sorted([y_i, y_j])
    x_j, x_i = sorted([x_i, x_j])
    if not ((x_i - x_j) > (y_i - y_j)):
        y_D = 1 / y_D
    first_term = 2 * np.pi * y_D * (1 / 3.0 - y_i / y_D + (y_i**2 + y_j**2) / (2 * y_D**2))

    tm = (
        np.cosh(m * np.pi * (y_D - np.abs(y_i - y_j))) + np.cosh(m * np.pi * (y_D - y_i - y_j))
    ) / np.sinh(m * np.pi * y_D)

    S1 = 2 * np.sum(tm / m * np.cos(m * np.pi * x_i) * np.cos(m * np.pi * x_j))
    tN = tm[-1]
    S2 = -tN / 2 * np.log(
        (1 - np.cos(np.pi * (x_i + x_j))) ** 2 + np.sin(np.pi * (x_i + x_j)) ** 2
    ) - tN / 2 * np.log((1 - np.cos(np.pi * (x_i - x_j))) ** 2 + np.sin(np.pi * (x_i - x_j)) ** 2)
    S3 = -2 * tN * np.sum(1 / m * np.cos(m * np.pi * x_i) * np.cos(m * np.pi * x_j))
    summed_term = S1 + S2 + S3

    return first_term + summed_term



"""Analyze waterfloods with capacitance-resistance models.

This is the central module in :code:`pywaterflood`, based around the:code:`CRM`
class, which implements the standard capacitance-resistance models. For most
cases, the best performance comes from selecting
:code:`CRM(primary=True, tau_selection="per-pair", constraints="up-to one")`.
If the data is too sparse, then change :code:`tau_selection` to "per-producer".

The base class assumes constant bottomhole pressures for the producing wells.
If you know the pressures for these wells or at least the trend, consider using
:code:`CrmCompensated`.

"""
from __future__ import annotations

import pickle
from pathlib import Path
from typing import Any

import _core
import numpy as np
import pandas as pd
from joblib import Parallel, delayed
from numpy.typing import NDArray
from scipy import optimize


def q_primary(
    production: NDArray, time: NDArray, gain_producer: float, tau_producer: float
) -> NDArray:
    """Calculate primary production contribution.

    Uses Arps equation with :math:`b=0`

    .. math::
        q_{p}(t) = q_i e^{-bt}

    Args
    ----------
    production : NDArray
        Production, size: Number of time steps
    time : NDArray
        Producing times to forecast, size: Number of time steps
    gain_producer : float
        Arps q_i factor
    tau_producer : float
        Arps time constant

    Returns
    ----------
    q_hat : NDArray
        Calculated production, size: Number of time steps
    """
    return _core.q_primary(production, time, gain_producer, tau_producer)


def q_CRM_perpair(injection: NDArray, time: NDArray, gains: NDArray, taus: NDArray) -> NDArray:
    """Calculate per injector-producer pair production.

    Runs for influences of each injector on one producer, assuming
    individual :code:`gain` and :code:`tau` for each pair

    Args
    ----------
    injection : NDArray
        Injected fluid, size: Number of time steps
    time : NDArray
        Producing times to forecast, size: Number of time steps
    gains : NDArray
        Connectivities between each injector and the producer,
        size: Number of injectors
    taus : NDArray
        Time constants between each injector and the producer,
        size: Number of injectors

    Returns
    ----------
    q_hat : NDArray
        Calculated production, size: Number of time steps
    """
    return _core.q_crm_perpair(injection, time, gains, taus)


def q_CRM_perproducer(injection: NDArray, time: NDArray, gain: NDArray, tau: float) -> NDArray:
    """Calculate per injector-producer pair production (simplified tank).

    Uses simplified CRMp model that assumes a single tau for each producer

    Args
    ----------
    injection : NDArray
        injected fluid in reservoir volumes, size: Number of time steps
    time : NDArray
        Producing times to forecast, size: Number of time steps
    gains : NDArray
        Connectivities between each injector and the producer
        size: Number of injectors
    tau : float
        Time constants all injectors and the producer

    Returns
    ----------
    q_hat : NDArray
        Calculated production, size: Number of time steps
    """
    tau2 = np.full(injection.shape[1], tau)
    return q_CRM_perpair(injection, time, gain, tau2)


def q_bhp(pressure_local: NDArray, pressure: NDArray, v_matrix: NDArray) -> NDArray:
    r"""Calculate the production effect from bottom-hole pressure variation.

    This looks like

    .. math::
        q_{BHP,j}(t_i) = \sum_{k} v_{kj}\left[ p_j(t_{i-1}) - p_k(t_i) \right]

    Args
    ----
    pressure_local : NDArray
        pressure for the well in question, shape: n_time
    pressure : NDArray
        bottomhole pressure, shape: n_time, n_producers
    v_matrix : NDArray
        connectivity between one producer and all producers, shape: n_producers

    Returns
    -------
    q : NDArray
        production from changing BHP
        shape: n_time
    """
    return _core.q_bhp(pressure_local, pressure, v_matrix)


def random_weights(n_prod: int, n_inj: int, axis: int = 0, seed: int | None = None) -> NDArray:
    """Generate random weights for producer-injector gains.

    Args
    ----
    n_i : int
    n_j : int
    axis : int, default is 0
    seed : int, default is None

    Returns
    -------
    gains_guess: NDArray
    """
    rng = np.random.default_rng(seed)
    limit = 10 * (n_prod if axis == 0 else n_inj)
    vec = rng.integers(0, limit, (n_prod, n_inj))
    axis_sum = vec.sum(axis, keepdims=True)
    return vec / axis_sum


class CRM:
    """A Capacitance Resistance Model history matcher.

    CRM uses a physics-inspired mass balance approach to explain production for
    waterfloods. It treats each injector-producer well pair as a system
    with mass input, output, and pressure related to the mass balance.
    Several versions exist. Select them from the arguments.

    Args
    ----------
    primary : bool
        Whether to model primary production (strongly recommended)
    tau_selection : str
        How many tau values to select
            - If 'per-pair', fit tau for each producer-injector pair
            - If 'per-producer', fit tau for each producer (CRMp model)
    constraints : str
        How to constrain the gains
            * If 'up-to one' (default), let gains vary from 0 (no connection) to 1 \
           (all injection goes to producer)
            * If 'positive', require each gain to be positive \
                (It is unlikely to go negative in real life)
            * If 'sum-to-one', require the gains for each injector to sum to one \
                            (all production accounted for)
            * If 'sum-to-one injector' (not implemented), require each injector's \
                gains to sum to one (all injection accounted for)

    Examples
    ----------
    crm = CRM(True, "per-pair", "up-to one")

    References
    ----------
    "A State-of-the-Art Literature Review on Capacitance Resistance Models for
    Reservoir Characterization and Performance Forecasting" - Wanderley de Holanda
    et al., 2018. https://www.mdpi.com/1996-1073/11/12/3368
    """

    def __init__(
        self,
        primary: bool = True,
        tau_selection: str = "per-pair",
        constraints: str = "positive",
    ):
        """Initialize CRM with appropriate settings."""
        if not isinstance(primary, bool):
            msg = "primary must be a boolean"
            raise TypeError(msg)
        self.primary = primary
        if constraints not in (
            "positive",
            "up-to one",
            "sum-to-one",
            "sum-to-one injector",
        ):
            msg = "Invalid constraints"
            raise ValueError(msg)
        self.constraints = constraints
        self.tau_selection = tau_selection
        if tau_selection == "per-pair":
            self.q_CRM = q_CRM_perpair
        elif tau_selection == "per-producer":
            self.q_CRM = q_CRM_perproducer
        else:
            msg = f'tau_selection must be one of ("per-pair","per-producer"), not {tau_selection}'
            raise ValueError(msg)

    def fit(
        self,
        production: NDArray,
        injection: NDArray,
        time: NDArray,
        initial_guess: NDArray = None,
        num_cores: int = 1,
        random: bool = False,
        **kwargs,
    ):
        """Build a CRM model from the production and injection data.

        Args
        ----------
        production : NDArray
            production rates for each time period,
            shape: (n_time, n_producers)
        injection : NDArray
            injection rates for each time period,
            shape: (n_time, n_injectors)
        time : NDArray
            relative time for each rate measurement, starting from 0,
            shape: (n_time)
        initial_guess : NDArray
            initial guesses for gains, taus, primary production contribution
            shape: (len(guess), n_producers)
        num_cores (int): number of cores to run fitting procedure on, defaults to 1
        random : bool
            whether to randomly initialize the gains
        **kwargs:
            keyword arguments to pass to scipy.optimize fitting routine

        Returns
        ----------
        self: trained model
        """
        _validate_inputs(production, injection, time)
        self.production = production
        self.injection = injection
        self.time = time

        if not initial_guess:
            initial_guess = self._get_initial_guess(random=random)
        bounds, constraints = self._get_bounds()

        def fit_well(production, x0):
            # residual is an L2 norm
            def residual(x, production):
                return sum(
                    (production - self._calculate_qhat(x, production, injection, time)) ** 2
                )

            return optimize.minimize(
                residual,
                x0,
                bounds=bounds,
                constraints=constraints,
                args=(production,),
                **kwargs,
            )

        if num_cores == 1:
            results = map(fit_well, self.production.T, initial_guess)
        else:
            results = Parallel(n_jobs=num_cores)(
                delayed(fit_well)(p, x0) for p, x0 in zip(self.production.T, initial_guess)
            )

        opts_perwell = [self._split_opts(r["x"]) for r in results]
        gains_perwell, tau_perwell, gains_producer, tau_producer = map(list, zip(*opts_perwell))

        self.gains: NDArray = np.vstack(gains_perwell)
        self.tau: NDArray = np.vstack(tau_perwell)
        self.gains_producer = np.array(gains_producer)
        self.tau_producer = np.array(tau_producer)
        return self

    def predict(self, injection=None, time=None, connections=None, production=None):
        """Predict production for a trained model.

        If the injection and time are not provided, this will use the training values

        Args
        ----------
        injection : Optional NDArray
            The injection rates to input to the system, shape (n_time, n_inj)
        time : Optional NDArray
            The timesteps to predict
        connections : Optional dict
            if present, the gains, tau, gains_producer, tau_producer
            matrices
        production : Optional NDArray
            The production (only takes first row to use for primary production decline)

        Returns
        ----------
        q_hat :NDArray
            The predicted values, shape (n_time, n_producers)
        """
        if connections is not None:
            gains = connections.get("gains", self.gains)
            tau = connections.get("tau", self.tau)
            gains_producer = connections.get("gains_producer", self.gains_producer)
            tau_producer = connections.get("tau_producer", self.tau_producer)
        else:
            gains = self.gains
            tau = self.tau
            gains_producer = self.gains_producer
            tau_producer = self.tau_producer
        if production is None:
            production = self.production
        n_producers = production.shape[1]

        if int(injection is None) + int(time is None) == 1:
            msg = "predict() takes 1 or 3 arguments, 2 given"
            raise TypeError(msg)
        if injection is None:
            injection = self.injection
        if time is None:
            time = self.time
        if time.shape[0] != injection.shape[0]:
            msg = "injection and time need same number of steps"
            raise ValueError(msg)

        q_hat = np.zeros((len(time), n_producers))
        for i in range(n_producers):
            q_hat[:, i] += q_primary(production[:, i], time, gains_producer[i], tau_producer[i])
            q_hat[:, i] += self.q_CRM(injection, time, gains[i, :], tau[i])
        return q_hat

    def set_rates(self, production=None, injection=None, time=None):
        """Set production and injection rates and time array.

        Args
        -----
        production : NDArray
            production rates with shape (n_time, n_producers)
        injection : NDArray
            injection rates with shape (n_time, n_injectors)
        time : NDArray
            timesteps with shape n_time
        """
        _validate_inputs(production, injection, time)
        if production is not None:
            self.production = production
        if injection is not None:
            self.injection = injection
        if time is not None:
            self.time = time

    def set_connections(self, gains=None, tau=None, gains_producer=None, tau_producer=None):
        """Set waterflood properties.

        Args
        -----
        gains : NDArray
            connectivity between injector and producer
            shape: n_gains, n_producers
        tau : NDArray
            time-constant for injection to be felt by production
            shape: either n_producers or (n_gains, n_producers)
        gains_producer : NDArray
            gain on primary production, shape: n_producers
        tau_producer : NDArray
            Arps time constant for primary production, shape: n_producers
        """
        if gains is not None:
            self.gains = gains
        if tau is not None:
            self.tau = tau
        if gains_producer is not None:
            self.gains_producer = gains_producer
        if tau_producer is not None:
            self.tau_producer = tau_producer

    def residual(self, production=None, injection=None, time=None):
        """Calculate the production minus the predicted production for a trained model.

        If the production, injection, and time are not provided, this will use the
         training values

        Args
        ----------
        production : NDArray
            The production rates observed, shape: (n_timesteps, n_producers)
        injection : NDArray
            The injection rates to input to the system,
            shape: (n_timesteps, n_injectors)
        time : NDArray
            The timesteps to predict

        Returns
        ----------
        residual : NDArray
            The true production data minus the predictions, shape (n_time, n_producers)
        """
        q_hat = self.predict(injection, time)
        if production is None:
            production = self.production
        return production - q_hat

    def to_excel(self, fname: str):
        """Write trained model to an Excel file.

        Args
        ----
        fname : str
            Excel file to write out

        """
        for x in ("gains", "tau", "gains_producer", "tau_producer"):
            if x not in self.__dict__:
                msg = "Model has not been trained"
                raise ValueError(msg)
        with pd.ExcelWriter(fname) as f:
            pd.DataFrame(self.gains).to_excel(f, sheet_name="Gains")
            pd.DataFrame(self.tau).to_excel(f, sheet_name="Taus")
            pd.DataFrame(
                {
                    "Producer gains": self.gains_producer,
                    "Producer taus": self.tau_producer,
                }
            ).to_excel(f, sheet_name="Primary production")

    def to_pickle(self, fname: str):
        """Write trained model to a pickle file.

        Args
        -----
        fname : str
            pickle file to write out
        """
        with Path(fname).open("wb") as f:
            pickle.dump(self, f)

    def _get_initial_guess(self, tau_selection: str | None = None, random=False):
        """Create initial guesses for the CRM model parameters.

        :meta private:

        Args
        ----------
        tau_selection : str, one of 'per-pair' or 'per-producer'
            sets whether to use CRM (per-pair) or CRMp model
        random : bool
            whether initial gains are randomly (true) or proportionally assigned
        Returns
        ----------
        x0 : NDArray
            Initial primary production gain, time constant and waterflood gains
            and time constants, as one long 1-d array
        """
        if tau_selection is not None:
            self.tau_selection = tau_selection

        n_inj = self.injection.shape[1]
        n_prod = self.production.shape[1]
        d_t = self.time[1] - self.time[0]

        axis = 1 if (self.constraints == "sum-to-one injector") else 0
        if random:
            rng = np.random.default_rng()
            gains_producer_guess1 = rng.random(n_prod)
            gains_guess1 = random_weights(n_prod, n_inj, axis)
        else:
            gains_unnormed = np.ones((n_prod, n_inj))
            gains_guess1 = gains_unnormed / np.sum(gains_unnormed, axis, keepdims=True)
            gains_producer_guess1 = np.ones(n_prod)
        tau_producer_guess1 = d_t * np.ones(n_prod)
        if self.tau_selection == "per-pair":
            tau_guess1 = d_t * np.ones((n_prod, n_inj))
        else:  # 'per-producer'
            tau_guess1 = d_t * np.ones((n_prod, 1))

        if self.primary:
            x0 = [
                np.concatenate(
                    [
                        gains_guess1[i, :],
                        tau_guess1[i, :],
                        gains_producer_guess1[[i]],
                        tau_producer_guess1[[i]],
                    ]
                )
                for i in range(n_prod)
            ]
        else:
            x0 = [np.concatenate([gains_guess1[i, :], tau_guess1[i, :]]) for i in range(n_prod)]
        return x0

    def _opt_numbers(self) -> tuple[int, int, int]:
        """Return the number of gains, taus, and primary production parameters to fit."""
        n_gains = self.injection.shape[1]
        n_tau = n_gains if self.tau_selection == "per-pair" else 1
        n_primary = 2 if self.primary else 0
        return n_gains, n_tau, n_primary

    def _get_bounds(self, constraints: str = "") -> tuple[tuple, tuple | dict]:
        """Create bounds for the model from initialized constraints."""
        if constraints:
            self.constraints = constraints

        n_inj = self.injection.shape[1]
        n = sum(self._opt_numbers())

        if self.constraints == "positive":
            bounds = ((0, np.inf),) * n
            constraints_optimizer = ()  # type: tuple | dict
        elif self.constraints == "sum-to-one":
            bounds = ((0, np.inf),) * n

            def constrain(x):
                x = x[:n_inj]
                return np.sum(x) - 1

            constraints_optimizer = {"type": "eq", "fun": constrain}
        elif self.constraints == "up-to one":
            lb = np.full(n, 0)
            ub = np.full(n, np.inf)
            ub[:n_inj] = 1
            bounds = tuple(zip(lb, ub))
            constraints_optimizer = ()
        elif self.constraints == "sum-to-one injector":
            msg = "sum-to-one injector is not implemented"
            raise NotImplementedError(msg)
        else:
            bounds = ((0, np.inf),) * n
            constraints_optimizer = ()
        return bounds, constraints_optimizer

    def _calculate_qhat(
        self,
        x: NDArray,
        production: NDArray,
        injection: NDArray,
        time: NDArray,
    ):
        gains, tau, gain_producer, tau_producer = self._split_opts(x)
        if self.primary:
            q_hat = q_primary(production, time, gain_producer, tau_producer)
        else:
            q_hat = np.zeros(len(time))

        q_hat += self.q_CRM(injection, time, gains, tau)
        return q_hat

    def _split_opts(self, x: NDArray):
        n_inj = self.injection.shape[1]

        gains = x[:n_inj]
        tau = x[n_inj : n_inj * 2] if self.tau_selection == "per-pair" else x[n_inj]
        if self.primary:
            gain_producer = x[-2]
            tau_producer = x[-1]
        else:
            gain_producer = 0
            tau_producer = 1
        if self.tau_selection == "per-pair":
            tau[tau < 1e-10] = 1e-10
        elif tau < 1e-10:
            tau = 1e-10
        if tau_producer < 1e-10:
            tau_producer = 1e-10
        return gains, tau, gain_producer, tau_producer


class CrmCompensated(CRM):
    """Bottom-hole pressure compensated CRM."""

    def fit(
        self,
        production: NDArray,
        pressure: NDArray,
        injection: NDArray,
        time: NDArray,
        initial_guess: NDArray = None,
        num_cores: int = 1,
        random: bool = False,
        **kwargs,
    ):
        """Fit a CRM model from the production, pressure, and injection data.

        Args
        ----------
        production : NDArray
            production rates for each time period,
            shape: (n_time, n_producers)
        pressure : NDArray
            average pressure for each producer for each time period,
            shape: (n_time, n_producers)
        injection : NDArray
            injection rates for each time period,
            shape: (n_time, n_injectors)
        time : NDArray
            relative time for each rate measurement, starting from 0,
            shape: (n_time)
        initial_guess : NDArray
            initial guesses for gains, taus, primary production
            contribution
            shape: (len(guess), n_producers)
        num_cores : int
            number of cores to run fitting procedure on, defaults to 1
        random : bool
            whether to randomly initialize the gains
        **kwargs:
            keyword arguments to pass to scipy.optimize fitting routine

        Returns
        ----------
        self: trained model
        """
        _validate_inputs(production, injection, time, pressure)
        self.production = production
        self.injection = injection
        self.time = time
        self.pressure = pressure

        if initial_guess is None:
            initial_guess = self._get_initial_guess(random=random)
        bounds, constraints = self._get_bounds()

        def fit_well(production, pressure_local, x0):
            # residual is an L2 norm
            def residual(x, production):
                return sum(
                    (
                        production
                        - self._calculate_qhat(
                            x, production, injection, time, pressure_local, pressure
                        )
                    )
                    ** 2
                )

            return optimize.minimize(
                residual,
                x0,
                bounds=bounds,
                constraints=constraints,
                args=(production,),
                **kwargs,
            )

        if num_cores == 1:
            results = map(fit_well, self.production.T, pressure.T, initial_guess)
        else:
            results = Parallel(n_jobs=num_cores)(
                delayed(fit_well)(prod, pressure, x0)
                for prod, pressure, x0 in zip(self.production.T, pressure.T, initial_guess)
            )

        opts_perwell = [self._split_opts(r["x"]) for r in results]
        gains_perwell, tau_perwell, gains_producer, tau_producer, gain_pressure = map(
            list, zip(*opts_perwell)
        )

        self.gains: NDArray = np.vstack(gains_perwell)
        self.tau: NDArray = np.vstack(tau_perwell)
        self.gains_producer = np.array(gains_producer)
        self.tau_producer = np.array(tau_producer)
        self.gain_pressure: NDArray = np.vstack(gain_pressure)
        return self

    def _calculate_qhat(
        self,
        x: NDArray,
        production: NDArray,
        injection: NDArray,
        time: NDArray,
        pressure_local: NDArray,
        pressure: NDArray,
    ):
        gains, tau, gain_producer, tau_producer, gain_pressure = self._split_opts(x)
        if self.primary:
            q_hat = q_primary(production, time, gain_producer, tau_producer)
        else:
            q_hat = np.zeros(len(time))

        q_hat += self.q_CRM(injection, time, gains, tau)
        q_hat += q_bhp(pressure_local, pressure, gain_pressure)
        return q_hat

    def _opt_numbers(self) -> tuple[int, int, int, int]:
        n_gain, n_tau, n_primary = super()._opt_numbers()
        return n_gain, n_tau, n_primary, self.production.shape[1]

    def _split_opts(self, x: NDArray) -> tuple[NDArray, NDArray, Any, Any, NDArray]:
        n_gains, n_tau, n_primary = self._opt_numbers()[:3]
        n_connectivity = n_gains + n_tau

        gains = x[:n_gains]
        tau = x[n_gains:n_connectivity]
        if self.primary:
            gain_producer = x[n_connectivity:][0]
            tau_producer = x[n_connectivity:][1]
        else:
            gain_producer = 0
            tau_producer = 1
        gain_pressure = x[n_connectivity + n_primary :]

        # boundary setting
        if self.tau_selection == "per-pair":
            tau[tau < 1e-10] = 1e-10
        elif tau < 1e-10:
            tau = 1e-10
        if tau_producer < 1e-10:
            tau_producer = 1e-10
        return gains, tau, gain_producer, tau_producer, gain_pressure

    def _get_initial_guess(self, tau_selection: str | None = None, random=False):
        """Make the initial guesses for the CRM model parameters.

        :meta private:

        Args
        ----------
        tau_selection : str, one of 'per-pair' or 'per-producer'
            sets whether to use CRM (per-pair) or CRMp model

        Returns
        ----------
        x0 : NDArray
            Initial primary production gain, time constant and waterflood gains
            and time constants, as one long 1-d array
        """
        guess = super()._get_initial_guess(tau_selection=tau_selection, random=random)
        _, _, _, n_pressure = self._opt_numbers()
        pressure_guess = np.ones(n_pressure)
        return [np.concatenate([guess[i], pressure_guess]) for i in range(len(guess))]


def _validate_inputs(
    production: NDArray | None = None,
    injection: NDArray | None = None,
    time: NDArray | None = None,
    pressure: NDArray | None = None,
) -> None:
    """Validate shapes and values of inputs.

    Args
    ----
    production : NDArray, optional
    injection : NDArray, optional
    time : NDArray, optional
    pressure : NDArray, optional

    Raises
    ------
    ValueError if timesteps don't match or production and pressure don't match
    """
    inputs = {
        "production": production,
        "injection": injection,
        "time": time,
        "pressure": pressure,
    }
    inputs = {key: val for key, val in inputs.items() if val is not None}
    # Shapes
    test_prod_inj_timesteps = production is not None and injection is not None
    if test_prod_inj_timesteps and (production.shape[0] != injection.shape[0]):
        msg = "production and injection do not have the same number of timesteps"
        raise ValueError(msg)
    if time is not None:
        for timeseries in inputs:
            if inputs[timeseries].shape[0] != time.shape[0]:
                msg = f"{timeseries} and time do not have the same number of timesteps"
                raise ValueError(msg)
    if production is not None:
        if (injection is not None) and (production.shape[0] != injection.shape[0]):
            msg = "production and injection do not have the same number of timesteps"
            raise ValueError(msg)
        if (pressure is not None) and (production.shape != pressure.shape):
            msg = "production and pressure are not of the same shape"
            raise ValueError(msg)
    if (
        (injection is not None)
        and (pressure is not None)
        and (injection.shape[0] != pressure.shape[0])
    ):
        msg = "injection and pressure do not have the same number of timesteps"
        raise ValueError(msg)
    # Values
    for timeseries in inputs:
        if np.any(np.isnan(inputs[timeseries])):
            msg = f"{timeseries} cannot have NaNs"
            raise ValueError(msg)
        if np.any(inputs[timeseries] < 0.0):
            msg = f"{timeseries} cannot be negative"
            raise ValueError(msg)



from __future__ import annotations

from itertools import product

import numpy as np
import pytest
from pywaterflood.crm import (
    CRM,
    CrmCompensated,
    q_bhp,
    q_CRM_perpair,
    q_CRM_perproducer,
    q_primary,
    random_weights,
)

primary = (True, False)
tau_selection = ("per-pair", "per-producer")
constraints = ("positive", "up-to one", "sum-to-one")
test_args = list(product(primary, tau_selection, constraints))
data_dir = "testing/data/"


@pytest.fixture()
def reservoir_simulation_data():
    injection = np.genfromtxt(data_dir + "injection.csv", delimiter=",")
    production = np.genfromtxt(data_dir + "production.csv", delimiter=",")
    time = np.genfromtxt(data_dir + "time.csv", delimiter=",")
    return injection, production, time


@pytest.fixture()
def trained_model(reservoir_simulation_data):
    def _trained_model(*args, **kwargs):
        crm = CRM(*args, **kwargs)
        crm.injection, crm.production, crm.time = reservoir_simulation_data
        crm.gains = np.genfromtxt(data_dir + "gains.csv", delimiter=",")
        if crm.tau_selection == "per-pair":
            crm.tau = np.genfromtxt(data_dir + "taus_per-pair.csv", delimiter=",")
        else:
            crm.tau = np.genfromtxt(data_dir + "taus.csv", delimiter=",")
        crm.gains_producer = np.genfromtxt(data_dir + "gain_producer.csv", delimiter=",")
        crm.tau_producer = np.genfromtxt(data_dir + "tau_producer.csv", delimiter=",")
        return crm

    return _trained_model


def test_q_primary(reservoir_simulation_data):
    "primary production."
    _, production, time = reservoir_simulation_data
    _, nprod = production.shape
    rng = np.random.default_rng(14_210)
    gain_producer = rng.uniform(0, 1, nprod)
    tau_producer = rng.uniform(1, 100, nprod)
    q = np.empty_like(production)
    for i in range(nprod):
        q[:, i] = q_primary(production[i], time, gain_producer[i], tau_producer[i])
    assert not np.isnan(q).flatten().any(), "No NaNs"
    pxg = production[0] * gain_producer
    summed_q = q.sum(0)
    assert np.allclose(pxg.argsort(), summed_q.argsort()), "Higher gains -> higher predictions"


def test_q_perpair(reservoir_simulation_data):
    "Secondary production."
    injection, production, time = reservoir_simulation_data
    _, n_prod = production.shape
    _, n_inj = injection.shape
    gains = random_weights(n_prod, n_inj, seed=42)
    rng = np.random.default_rng(42)
    taus = rng.uniform(1, 40, gains.shape)
    q_1 = np.empty_like(production)
    q_2 = np.empty_like(production)
    for i in range(n_prod):
        q_1[:, i] = q_CRM_perpair(injection, time, gains[i], taus[i])
        q_2[:, i] = q_CRM_perpair(injection, time, 0.5 * gains[i], taus[i])
    assert not np.isnan(q_1).flatten().any(), "No NaNs"
    assert not np.isnan(q_1).flatten().any(), "No NaNs"
    assert np.allclose(q_1, 2.0 * q_2), "Double gains -> double prod"


def test_q_perproducer(reservoir_simulation_data):
    "Secondary production."
    injection, production, time = reservoir_simulation_data
    _, n_prod = production.shape
    _, n_inj = injection.shape
    gains = random_weights(n_prod, n_inj, seed=42)
    rng = np.random.default_rng(42)
    taus = rng.uniform(1, 40, n_prod)
    q_1 = np.empty_like(production)
    q_2 = np.empty_like(production)
    for i in range(n_prod):
        q_1[:, i] = q_CRM_perproducer(injection, time, gains[i], taus[i])
        q_2[:, i] = q_CRM_perproducer(injection, time, 0.5 * gains[i], taus[i])
    assert not np.isnan(q_1).flatten().any(), "No NaNs"
    assert not np.isnan(q_1).flatten().any(), "No NaNs"
    assert np.allclose(q_1, 2.0 * q_2), "Double gains -> double prod"


def test_q_bhp_nonan(reservoir_simulation_data):
    "q_BHP functions"
    production = reservoir_simulation_data[1]
    _, nprod = production.shape
    pressure = production.copy()
    rng = np.random.default_rng(42)
    producer_gains = rng.normal(0, 1, nprod)
    q = q_bhp(pressure[:, 0], pressure, producer_gains)
    assert not np.isnan(q).any(), "No NaNs among results"


def test_q_bhp():
    """Test bottomhole pressure influencing production."""
    n_time = 5
    n_prod = 2
    pressure = np.ones((n_time, n_prod))
    producer_gains = np.random.rand(n_prod)
    q = q_bhp(pressure[:, 0], pressure, producer_gains)
    assert np.allclose(q, 0), "no pressure change -> no prod"
    pressure[-1] = 0
    producer_gains = np.ones(n_prod)
    q = q_bhp(pressure[:, 0], pressure, producer_gains)
    assert np.allclose(q[-1], 2), "pressure drop -> increase production"
    pressure[-1] = 2
    producer_gains = np.ones(n_prod)
    q = q_bhp(pressure[:, 0], pressure, producer_gains)
    assert np.allclose(q[-1], -2), "pressure increase -> drop production"


@pytest.mark.parametrize("primary,tau_selection,constraints", test_args)
class TestInstantiate:
    def test_init(self, primary, tau_selection, constraints):
        CRM(primary, tau_selection, constraints)

    def test_init_fails(self, primary, tau_selection, constraints):
        with pytest.raises(TypeError):
            CRM(primary="yes", tau_selection=tau_selection, constraints=constraints)
        with pytest.raises(ValueError, match="constraints"):
            CRM(primary=primary, tau_selection=tau_selection, constraints="negative")
        with pytest.raises(ValueError, match="tau"):
            CRM(primary=primary, tau_selection="per-Bob", constraints=constraints)


@pytest.mark.parametrize("tau_selection", tau_selection)
@pytest.mark.parametrize("primary", primary)
class TestPredict:
    def test_predict(self, reservoir_simulation_data, trained_model, primary, tau_selection):
        injection, production, time = reservoir_simulation_data
        crm = trained_model(primary=primary, tau_selection=tau_selection)

        prediction1 = crm.predict()
        prediction2 = crm.predict(injection, time)
        prediction3 = crm.predict(connections={"gains": crm.gains})
        assert prediction1 == pytest.approx(prediction2, abs=1.0)
        assert prediction2 == pytest.approx(prediction3, abs=1)

        primary_str = "primary" if primary else "noprimary"

        assert prediction1 == pytest.approx(
            np.genfromtxt(
                f"{data_dir}prediction_{primary_str}_{tau_selection}.csv", delimiter=","
            ),
            abs=5.0,
            rel=1e-2,
        )

    def test_predict_fails(self, reservoir_simulation_data, trained_model, primary, tau_selection):
        injection, production, time = reservoir_simulation_data
        crm = trained_model(primary, tau_selection)
        with pytest.raises(TypeError, match="arguments"):
            crm.predict(injection)
        with pytest.raises(ValueError, match="number of steps"):
            crm.predict(injection, time[:-1])

    def test_set_connections(
        self, reservoir_simulation_data, trained_model, primary, tau_selection
    ):
        injection, production, time = reservoir_simulation_data
        crm = trained_model(primary=primary, tau_selection=tau_selection)
        crm2 = CRM(primary=primary, tau_selection=tau_selection)
        crm2.set_connections(
            gains=crm.gains,
            tau=crm.tau,
            gains_producer=crm.gains_producer,
            tau_producer=crm.tau_producer,
        )
        crm2.set_connections()  # no-op, hopefully
        prediction2 = crm2.predict(injection, time, production=production)
        assert crm.predict(injection, time) == pytest.approx(prediction2)
        assert production.shape == prediction2.shape

    def test_residual(self, reservoir_simulation_data, trained_model, primary, tau_selection):
        injection, production, time = reservoir_simulation_data
        crm = trained_model(primary=primary, tau_selection=tau_selection)
        q_hat = crm.predict(injection, time)
        resid1 = production - q_hat
        resid2 = crm.residual(production, injection, time)
        assert resid1 == pytest.approx(resid2)
        assert resid1.shape == (len(time), production.shape[-1])


@pytest.mark.parametrize("primary,tau_selection,constraints", test_args)
class TestFit:
    def test_validate_timeseries(
        self, reservoir_simulation_data, primary, tau_selection, constraints
    ):
        injection, production, time = reservoir_simulation_data
        crm = CRM(primary, tau_selection, constraints)
        with pytest.raises(ValueError, match="same number"):
            crm.set_rates(production[:-5], injection, time)
        with pytest.raises(ValueError, match="same number"):
            crm.set_rates(production, injection[:-5], time)
        with pytest.raises(ValueError, match="same number"):
            crm.set_rates(production, injection, time[:-5])
        prod_bad = production.copy()
        prod_bad[4, 0] = -1
        with pytest.raises(ValueError, match="negative"):
            crm.set_rates(prod_bad, injection, time)
        prod_bad[4, 0] = np.nan
        with pytest.raises(ValueError, match="NaN"):
            crm.set_rates(prod_bad, injection, time)
        inj_bad = injection.copy()
        inj_bad[2, 2] = -1
        with pytest.raises(ValueError, match="negative"):
            crm.set_rates(production, inj_bad, time)
        inj_bad[2, 2] = np.nan
        with pytest.raises(ValueError, match="NaN"):
            crm.set_rates(production, inj_bad, time)
        time_bad = time.copy()
        time_bad[0] = -1
        with pytest.raises(ValueError, match="negative"):
            crm.set_rates(time=time_bad)
        time_bad[0] = np.nan
        with pytest.raises(ValueError, match="NaN"):
            crm.set_rates(time=time_bad)

    def test_fit_fails(self, reservoir_simulation_data, primary, tau_selection, constraints):
        crm = CRM(primary, tau_selection, constraints)
        injection, production, time = reservoir_simulation_data

        with pytest.raises(TypeError, match="missing .* positional argument"):
            crm.fit(production)
        with pytest.raises(TypeError, match="missing .* positional argument"):
            crm.fit(production, injection)
        with pytest.raises(ValueError, match="same number of timesteps"):
            crm.fit(production[:-1], injection, time)
        with pytest.raises(ValueError, match="same number of timesteps"):
            crm.fit(production, injection[:-1], time)
        with pytest.raises(ValueError, match="same number of timesteps"):
            crm.fit(production, injection, time[:-1])
        crm.constraints = "sum-to-one injector"
        with pytest.raises(NotImplementedError):
            crm.fit(production, injection, time)

    @pytest.mark.slow()
    @pytest.mark.parametrize("random", [True, False])
    def test_fit_serial(
        self, reservoir_simulation_data, primary, tau_selection, constraints, random
    ):
        injection, production, time = reservoir_simulation_data
        crm = CRM(primary, tau_selection, constraints)
        crm.fit(
            production,
            injection,
            time,
            num_cores=1,
            options={"maxiter": 3},
            random=random,
        )

    @pytest.mark.parametrize("random", [True, False])
    def test_fit_parallel(
        self, reservoir_simulation_data, primary, tau_selection, constraints, random
    ):
        injection, production, time = reservoir_simulation_data
        crm = CRM(primary, tau_selection, constraints)
        crm.fit(
            production,
            injection,
            time,
            num_cores=4,
            options={"maxiter": 3},
            random=random,
        )

    @pytest.mark.slow()
    @pytest.mark.parametrize("random", [True, False])
    def test_fit_initial_guess(
        self, reservoir_simulation_data, primary, tau_selection, constraints, random
    ):
        injection, production, time = reservoir_simulation_data
        crm = CRM(primary, tau_selection, constraints)
        crm.set_rates(production, injection, time)
        crm.set_rates()  # no-op
        x0 = crm._get_initial_guess(tau_selection)
        crm.fit(
            production,
            injection,
            time,
            random=random,
            initial_guess=x0,
            options={"maxiter": 3},
        )


@pytest.mark.parametrize("primary", primary)
class TestExport:
    def test_to_excel(self, trained_model, primary, tmpdir):
        crm = trained_model(primary)
        crm.to_excel(tmpdir + "/test.xlsx")

    def test_to_excel_fails(self, trained_model, primary, tmpdir):
        crm = trained_model(primary)
        with pytest.raises(TypeError, match="missing .* argument"):
            crm.to_excel()
        crm2 = CRM(primary)
        with pytest.raises(ValueError, match="Model has not been trained"):
            crm2.to_excel(tmpdir + "/test.xlsx")

    def test_to_pickle(self, trained_model, primary, tmpdir):
        crm = trained_model(primary)
        crm.to_pickle(tmpdir + "/test.pkl")

    def test_to_pickle_fails(self, trained_model, primary):
        crm = trained_model(primary)
        with pytest.raises(TypeError, match="missing .* argument"):
            crm.to_pickle()


@pytest.mark.parametrize("primary,tau_selection,constraints", test_args)
class TestBhp:
    def test_init_bhp(self, primary, tau_selection, constraints):
        crm = CrmCompensated(primary, tau_selection, constraints)
        assert crm.primary == primary

    @pytest.mark.parametrize("random", [True, False])
    @pytest.mark.parametrize("num_cores", [1, 4])
    @pytest.mark.parametrize("initial_guess", [True, None])
    def test_fit(
        self,
        reservoir_simulation_data,
        primary,
        tau_selection,
        constraints,
        random,
        num_cores,
        initial_guess,
    ):
        injection, production, time = reservoir_simulation_data
        crm = CrmCompensated(primary, tau_selection, constraints)
        crm.set_rates(production, injection, time)
        pressure = np.ones_like(production)
        if initial_guess:
            initial_guess = crm._get_initial_guess(random=random)
        crm.fit(
            production,
            pressure,
            injection,
            time,
            initial_guess,
            num_cores=num_cores,
            options={"maxiter": 3},
            random=random,
        )
        prediction1 = crm.predict()
        prediction2 = crm.predict(injection, time)
        assert prediction1 == pytest.approx(prediction2)



"""Regenerate results.

This is for regenerating the results that are tested against during test_predict. Every once in
a while it makes sense to regenerate it, but be careful.

functions
--------
test_predict

use
----
run `python testing/regenerate_results.py` from the project directory
"""
from __future__ import annotations

import numpy as np
from pywaterflood import CRM

# from pywaterflood.crm import CrmCompensated, q_bhp

data_dir = "testing/data/"


def reservoir_simulation_data():
    injection = np.genfromtxt(data_dir + "injection.csv", delimiter=",")
    production = np.genfromtxt(data_dir + "production.csv", delimiter=",")
    time = np.genfromtxt(data_dir + "time.csv", delimiter=",")
    return injection, production, time


def _trained_model(*args, **kwargs):
    crm = CRM(*args, **kwargs)
    crm.injection, crm.production, crm.time = reservoir_simulation_data()
    crm.gains = np.genfromtxt(data_dir + "gains.csv", delimiter=",")
    if crm.tau_selection == "per-pair":
        crm.tau = np.genfromtxt(data_dir + "taus_per-pair.csv", delimiter=",")
    else:
        crm.tau = np.genfromtxt(data_dir + "taus.csv", delimiter=",")
    crm.gains_producer = np.genfromtxt(data_dir + "gain_producer.csv", delimiter=",")
    crm.tau_producer = np.genfromtxt(data_dir + "tau_producer.csv", delimiter=",")
    return crm


def test_predict(primary, tau_selection):
    """Predictions for test cases."""
    crm = _trained_model(primary=primary, tau_selection=tau_selection)
    return crm.predict()


if __name__ == "__main__":
    print("Regenerating prediction files")  # noqa: T201
    primary = (True, False)
    tau_selection = ("per-pair", "per-producer")
    for p in primary:
        for t in tau_selection:
            prediction = test_predict(p, t)
            primary_str = "primary" if p else "noprimary"
            np.savetxt(f"{data_dir}prediction_{primary_str}_{t}.csv", prediction, delimiter=",")
    print("Finished")  # noqa: T201



"""pywaterflood: A package for explaining and predicting waterflood performance.

Provides
--------
CRM : a class for performing capacitance resistance modeling
calc_gains_homogeneous : a function using the multiwell productivity index
translate_location : a function for moving things from the real world
  into dimensionless-land
"""
from __future__ import annotations

import sys

from pywaterflood.crm import CRM, CrmCompensated
from pywaterflood.multiwellproductivity import (
    calc_gains_homogeneous,
    translate_locations,
)

pyversion = sys.version_info
if pyversion.major == 3 and pyversion.minor > 7:
    from importlib.metadata import version

    __version__ = version("pywaterflood")
del sys, pyversion

__all__ = ["CRM", "CrmCompensated", "calc_gains_homogeneous", "translate_locations"]



from __future__ import annotations

from numpy.typing import NDArray

def q_primary(
    production: NDArray, time: NDArray, gain_producer: float, tau_producer: float
) -> NDArray: ...
def q_crm_perpair(injection: NDArray, time: NDArray, gains: NDArray, taus: NDArray) -> NDArray: ...
def q_bhp(pressure_local: NDArray, pressure: NDArray, v_matrix: NDArray) -> NDArray: ...
def calc_A_ij(x_i: float, y_i: float, x_j: float, y_j: float, y_d: float, m: NDArray) -> float: ...



"""Nox sessions for linting, docs, and testing."""
from __future__ import annotations

import argparse
import os
import shutil
from pathlib import Path

import nox

DIR = Path(__file__).parent.resolve()

nox.options.sessions = ["lint", "tests"]


@nox.session
def lint(session: nox.Session) -> None:
    """Run the linter.

    Includes all the pre-commit checks on all the files.
    """
    session.install("pre-commit")
    session.run("pre-commit", "run", "--all-files", *session.posargs)


@nox.session
def tests(session: nox.Session) -> None:
    """Run the unit and regular tests."""
    session.install("maturin")
    # You have to run `maturin develop` to avoid this: https://github.com/PyO3/maturin/issues/490
    session.run("maturin", "develop", "--release", "--extras=test")
    session.run(
        "pytest",
        "--cov=pywaterflood",
        "--cov-append",
        "--cov-report=xml",
        *session.posargs,
    )


@nox.session(reuse_venv=True)
def docs(session: nox.Session) -> None:
    """
    Build the docs. Pass "--serve" to serve. Pass "-b linkcheck" to check links.
    """

    parser = argparse.ArgumentParser()
    parser.add_argument("--serve", action="store_true", help="Serve after building")
    parser.add_argument("-b", dest="builder", default="html", help="Build target (default: html)")
    args, posargs = parser.parse_known_args(session.posargs)

    if args.builder != "html" and args.serve:
        session.error("Must not specify non-HTML builder with --serve")

    extra_installs = ["sphinx-autobuild"] if args.serve else []

    session.install("-e.[docs]", *extra_installs)
    session.chdir("docs")

    if args.builder == "linkcheck":
        session.run("sphinx-build", "-b", "linkcheck", ".", "_build/linkcheck", *posargs)
        return

    shared_args = (
        "-n",  # nitpicky mode
        "-T",  # full tracebacks
        f"-b={args.builder}",
        ".",
        f"_build/{args.builder}",
        *posargs,
    )

    if args.serve:
        session.run("sphinx-autobuild", *shared_args)
    else:
        session.run("sphinx-build", "--keep-going", *shared_args)


@nox.session
def build(session: nox.Session) -> None:
    """Build an SDist and a wheel."""
    build_p = DIR.joinpath("build")
    if build_p.exists():
        shutil.rmtree(build_p)

    session.install("build")
    session.run("python", "-m", "build", "-o", "build")


@nox.session
def paper(session: nox.Sesson) -> None:
    """Build the JOSS paper draft."""
    paper_dir = DIR.joinpath("paper")
    session.run(
        "docker",
        "run",
        "--rm",
        "--volume",
        f"{paper_dir}:/data",
        "--user",
        f"{os.getuid()}:{os.getgid()}",
        "--env",
        "JOURNAL=joss",
        "openjournals/inara",
        external=True,
    )
